from typing import List, Union

from MD6.hash.generator import HashGenerator


class Hash:
    __byte: List[int]
    hash_generator = HashGenerator()

    def __init__(self, data: Union[str, bytes], size: int, key: str, levels: int):
        self.__byte = self.hash_generator(data, size, key, levels)

    def __str__(self):
        return ''.join(["%02x" % i for i in self.__byte])

    def raw(self):
        return ''.join((chr(i) for i in self.__byte))



from typing import Optional, Union

from MD6.hash.hash import Hash


class MD6:
    __size: int
    __key: str
    __levels: int

    def __init__(self,
                 size: Optional[int] = None,
                 key: Optional[str] = None,
                 levels: Optional[int] = None):
        self.__size = 256 if size is None else size
        self.__key = "" if key is None else key
        self.__levels = 64 if levels is None else levels

    def __call__(self, data: Union[str, bytes]):
        return Hash(data, self.__size, self.__key, self.__levels)



import binascii
from typing import List, Union

from MD6.hash.calc import HashCalc


class HashGenerator:
    def __call__(self, data: Union[str, bytes], size: int, key: str, levels: int) -> List[int]:
        data = self.__byte_convert(data)
        key = self.__byte_convert(key)
        if size <= 0:
            size = 1
        elif size > 512:
            size = 512

        return self.__hash(data, size, key, levels)

    def __hash(self, data: List[int], size: int, key: List[int], levels: int):
        hashCalc = HashCalc(data, size, key, levels)
        d, M = hashCalc(data)
        return self.__crop(d, M, True)

    @staticmethod
    def __byte_convert(input_str: Union[str, bytes]) -> List[int]:
        input_str: bytes = binascii.hexlify(input_str.encode("ascii") if isinstance(input_str, str) else input_str)
        o_byte = [int(input_str[i:i + 2], 16) for i in range(0, len(input_str), 2)]

        return o_byte

    def __crop(self, size: int, data: List[int], right: bool):
        length = int((size + 7) / 8)
        remain = size % 8

        data = data[len(data) - length:] if right else data[:length]
        if remain > 0:
            data[length - 1] &= (0xff << (8 - remain)) & 0xff

        return data



from typing import List, Tuple

from MD6.constants import Constants


class HashCalc:
    @staticmethod
    def __to_word(i_byte: List[int]):
        return [((i_byte[i + 0] & 0xff) << 56) |
                ((i_byte[i + 1] & 0xff) << 48) |
                ((i_byte[i + 2] & 0xff) << 40) |
                ((i_byte[i + 3] & 0xff) << 32) |
                ((i_byte[i + 4] & 0xff) << 24) |
                ((i_byte[i + 5] & 0xff) << 16) |
                ((i_byte[i + 6] & 0xff) << 8) |
                ((i_byte[i + 7] & 0xff) << 0) for i in range(0, len(i_byte), 8)]

    @staticmethod
    def __from_word(i_word: List[int]):
        return [(i_word[i] >> j * 8) & 0xff for i in range(len(i_word)) for j in range(7, -1, -1)]

    def __init__(self, data: List[int], size: int, key: List[int], levels: int):
        K = key[:64]
        k = len(K)
        self.r = max(80 if k else 0, 40 + int(size / 4))
        self.K = self.__to_word(K + [0x00 for i in range(64 - len(K))])
        self.k = k
        self.ell = 0
        self.L = levels
        self.d = size

    def par(self, M: List[int]):
        P = 0
        B = []
        C = []
        z = 0 if len(M) > Constants.b else 1

        while len(M) < 1 or (len(M) % Constants.b) > 0:
            M.append(0x00)
            P += 8

        M = self.__to_word(M)
        while len(M) > 0:
            B.append(M[:int(Constants.b / 8)])
            M = M[int(Constants.b / 8):]
        i = 0
        p = 0
        l = len(B)

        while i < l:
            p = P if i == (len(B) - 1) else 0
            C += self.__mid(B[i], [], i, p, z)

            i += 1
            p = 0

        return self.__from_word(C)

    def seq(self, M: List[int]):
        P = 0
        B = []
        C = [0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0]

        while len(M) < 1 or (len(M) % (Constants.b - Constants.c)) > 0:
            M.append(0x00)
            P += 8

        M = self.__to_word(M)

        while len(M) > 0:
            B.append(M[:int((Constants.b - Constants.c) / 8)])
            M = M[int((Constants.b - Constants.c) / 8):]

        i = 0
        p = 0
        l = len(B)

        while i < l:
            p = P if i == (len(B) - 1) else 0
            z = 1 if i == (len(B) - 1) else 0
            C = self.__mid(B[i], C, i, p, z)

            i += 1
            p = 0

        return self.__from_word(C)

    def __f(self, N: List[int]):
        S = Constants.S0
        A = list(N)
        i = Constants.n

        for j in range(self.r):
            for s in range(16):
                x = S
                x ^= A[i + s - Constants.t[5]]
                x ^= A[i + s - Constants.t[0]]
                x ^= A[i + s - Constants.t[1]] & A[i + s - Constants.t[2]]
                x ^= A[i + s - Constants.t[3]] & A[i + s - Constants.t[4]]
                x ^= x >> Constants.rs[s]

                A += [0x00 for i in range(i + s - len(A) + 1)]
                A[i + s] = x ^ ((x << Constants.ls[s]) & 0xffffffffffffffff)

            S = (((S << 1) & 0xffffffffffffffff) ^ (S >> 63)) ^ (S & Constants.Sm)
            i += 16

        return A[(len(A) - 16):]

    def __mid(self, B: List[int], C: List[int], i: int, p: int, z: int):
        return self.__f(Constants.Q + self.K +
                        [((self.ell & 0xff) << 56) | i & 0xffffffffffffff,
                         ((self.r & 0xfff) << 48) | ((self.L & 0xff) << 40) | ((z & 0xf) << 36) | (
                                 (p & 0xffff) << 20) | (
                                 (self.k & 0xff) << 12) | (self.d & 0xfff)] +
                        C + B)

    def __call__(self, M: List[int]) -> Tuple[int, List[int]]:
        while True:
            self.ell += 1
            M = self.seq(M) if self.ell > self.L else self.par(M)

            if len(M) == Constants.c:
                break
        return self.d, M



from setuptools import setup, find_packages

setup(
    name='pymd6',
    packages=find_packages(),
    version='0.1.0',
    description='Simple library to calculate MD6 checksum',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author="Pokusaev Andrey",
    license="MIT",
    author_email='pockusaevandrey@yandex.ru',
    url='https://github.com/PockusaevAndrey/pymd6',
    keywords=['MD6'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)
